###############################################################################
#
# Function library for IME-based studies
#
###############################################################################

# Copyright (C) 2007-2008 Ian Korf
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

package IME;
use strict;
use warnings;

sub read_imeter_parameter_file {
	my ($file) = @_;

	my $info;
	my %model;
	my $wordsize;
	open(IN, $file) or die;
	while (<IN>) {
		if (/^#(.+)/) {
			 $info = $1 if not defined $info;
			 next;
		}
		my ($word, $score) = split;
		$word = uc $word;
		$model{$word} = $score;
		$wordsize = length($word) unless defined $wordsize;
	}
	close IN;
	$info = "" unless defined $info;
	
	return \%model, $wordsize, $info;	
}

sub write_imeter_parameter_file {
	my ($file, $model, $info) = @_;
	
	open(OUT, ">$file") or die;
	print OUT "# $info\n";
	foreach my $word (sort {$model->{$b} <=> $model->{$a}} keys %$model) {
		print OUT $word, "\t", $model->{$word}, "\n";
	}
	close OUT;
}

sub score_intron {
	my ($model, $w, $seq) = @_;
	
	$seq = uc $seq;
	my $score = 0;
	for (my $i = 0; $i < length($seq) - $w +1; $i++) {
		my $word = substr($seq, $i, $w);
		next unless exists $model->{$word}; # ignore ambiguities
		$score += $model->{$word};
	}
	
	return $score;
}

sub read_exon_intron_file {
	my ($file, $clip_don, $clip_acc, $min_intron, $max_intron) = @_;
	
	my %kill; # too long or too short
	my @Intron;
	my $id;
	my @data;
	my %data;
	my $count;
	my $distance;
	open(IN, $file) or die;
	while (<IN>) {
		if (/^>/) {
			($id, @data) = split;
			$id = substr($id, 1);
			%data = ();
			foreach my $item (@data) {
				my ($key, $value) = $item =~ /^(\S+)=(\S+)/;
				$data{$key} = $value;
			}
			$count = 1;
			$distance = 0;
		} elsif (/^[acgtn]/) {
			chomp;
			my %clone = (
				ordinal => $count++,
				gene => $id,
				seq => $_,
				distance => $distance,
			);
			foreach my $k (keys %data) {
				$clone{$k} = $data{$k};
			}
			push @Intron, \%clone;
			$distance += length($_);
			
			# length filter (applied to whole gene later)
			if (length($clone{seq}) < $min_intron or
			    length($clone{seq}) > $max_intron) {$kill{$clone{gene}} = 1}
		} elsif (/^[ACGTN]/) {
			$distance += length($_) -1;
		} elsif (/^\s+$/) {
			next;
		} else {
			die "unexpected input";
		}
			
	}
	close IN;
		
	my @prune;
	foreach my $intron (@Intron) {
		push @prune, $intron unless exists $kill{$intron->{gene}};
	}
		
	# clip acceptor and donor
	foreach my $intron (@prune) {
		$intron->{seq} = uc substr($intron->{seq}, $clip_don,
			length($intron->{seq}) - $clip_don - $clip_acc);
	}
	
	return \@prune;
}

sub train_imeter {
	my ($plus, $minus, $w) = @_; # refs to arrays of introns, word size
	
	my $f1 = freq_table($plus, $w);
	my $f2 = freq_table($minus, $w);
	my %lod;
	foreach my $word (keys %$f1) {
		$lod{$word} = log($f1->{$word}/$f2->{$word}) / log(2);
	}
	
	return \%lod;
}

sub freq_table {
	my ($introns, $w) = @_;
	
	# counts
	my $table = blank_table($w, 1); # 1 pseudocount
	foreach my $intron (@$introns) {
		for (my $i = 0; $i < length($intron->{seq}) - $w +1; $i++) {
			my $word = substr($intron->{seq}, $i, $w);
			next unless exists $table->{$word};
			$table->{$word}++;
		}
	}
	
	# convert to freq
	my $freq = count_to_freq($table);
	return $freq;
	
}


sub kl_distance {
	my ($f1, $f2) = @_;
	
	die "Unequal keys" unless keys %$f1 == keys %$f2;
	
	my $D = 0;
	foreach my $k (keys %$f1) {
		die "Zero not allowed" if $f1->{$k} == 0;
		die "Zero not allowed" if $f2->{$k} == 0;
		$D += $f1->{$k} * log($f1->{$k} / $f2->{$k});
	}
	
	return $D;
}

sub count_to_freq {
	my ($hashref) = @_;
	
	my $total = 0;
	foreach my $k (keys %$hashref) {
		$total += $hashref->{$k};
	}
	
	my %freq;
	foreach my $k (keys %$hashref) {
		$freq{$k} = $hashref->{$k} / $total;
	}
	
	return \%freq;
}

sub entropy_from_counts {
	my ($hashref) = @_;
	
	my $total = 0;
	foreach my $symbol (keys %$hashref) {
		$total += $hashref->{$symbol};
	}
	
	my %freq;
	foreach my $symbol (keys %$hashref) {
		$freq{$symbol} = $hashref->{$symbol} / $total;
	}
	
	return entropy(\%freq);
}

sub entropy {
	my ($hashref) = @_;
	
	my $H = 0;
	foreach my $symbol (keys %$hashref) {
		die "Zero not allowed" unless $hashref->{$symbol};
		$H += $hashref->{$symbol} * log($hashref->{$symbol});
	}
	
	return -$H;
}

sub blank_table {
	my ($mer, $pseudo) = @_;
	
	my $size = 4 ** $mer;
	$pseudo = 0 if not defined $pseudo;
	
	my %H;
	for (my $i = 0; $i < $size; $i++) {
		$H{number_to_dna($i, $mer)} = $pseudo;
	}
	
	return \%H;
}

sub number_to_dna {
	my ($n, $m) = @_;
	
	my $s = "";
	for (my $i = 0; $i < $m; $i++) {
		my $max = 4 ** ($m -$i -1);
		if ($max > $n) {
			$s .= 'A';
		} else {
			my $v = int ($n / $max);
			my $c;
			if    ($v == 0) {$c = 'A'}
			elsif ($v == 1) {$c = 'C'}
			elsif ($v == 2) {$c = 'G'}
			elsif ($v == 3) {$c = 'T'}
			else {die "out of bounds $v"}
			$s .= $c;
			$n -= $v * $max;
		}
	}
	
	return $s;
}

sub dna_to_number {
	my ($seq) = @_;
	
	my $total = 0;
	for (my $i = 1; $i <= length($seq); $i++) {
		my $nt = substr($seq, -$i, 1);
		my $v;
		if    ($nt eq 'A') {$v = 0}
		elsif ($nt eq 'C') {$v = 1}
		elsif ($nt eq 'G') {$v = 2}
		elsif ($nt eq 'T') {$v = 3}
		my $q = $v * (4 ** ($i -1));
		$total += $q;
	}
	
	return $total;
}

sub gc_composition {
	my ($seq) = @_;
	my $gc = 0;
	my $tot = 0;
	for (my $i = 0; $i < length($seq); $i++) {
		my $nt = substr($seq, $i, 1);
		if    ($nt =~ /[GgCc]/) {$gc++; $tot++}
		elsif ($nt =~ /[AaTt]/) {$tot++}
		# ignore non-[ACGT]
	}
	return $gc/$tot;
}

sub shuffle {
	my ($seq) = @_;
	my @seq = split(//, $seq);
	my $shuff = "";
	while (@seq) {
		$shuff .= splice(@seq, rand(@seq), 1);
	}
	return $shuff;
}

1;

__END__


Parameter files:

	* comments lines begin with #
	* first line should be a comment with information on how constructed
	* file data is a list of words and log2-odds scores
	* all words must be equal length
	* all words must be present for a particular word length
	* only {A, C, G, T} are allowed
